#include "SampleDecoder.h"

SolutionDecoder::SolutionDecoder() { }

SolutionDecoder::~SolutionDecoder() { }

ScenarioDecoder::ScenarioDecoder() { }

ScenarioDecoder::~ScenarioDecoder() { }

//Problem-dependent
std::vector<std::vector<double>> matrixF(int MAX_T, std::vector<std::vector<double>> popx, std::vector<std::vector<double>> pops) {
	std::vector<std::vector<double>> F(popx.size(), std::vector<double>(pops.size(), 0.0));


	#ifdef _OPENMP
		#pragma omp parallel for num_threads(MAX_T) 
	#endif
	for (int k = 0; k < popx.size()*pops.size(); k++) {
		int i = floor((double)(k) / (double)(pops.size())); //solution
		int j = k - i * pops.size();	//scenario
		

		//DECODE Solution

		std::vector< std::pair< double, unsigned > > ranking(popx[i].size());
		for (unsigned g = 0; g < popx[i].size(); ++g) {
			ranking[g] = std::pair< double, unsigned >(popx[i][g], g);
		}
		std::sort(ranking.begin(), ranking.end());

		//DECODE Scenario
		
		double auxsum = 0.0;
		for (unsigned g = 0; g < pops[j].size(); ++g) { 
			auxsum += pops[j][g];
		}
		
		//Calculate F value
		double finalval = 0.0;
		int genestoconsider = floor((auxsum / (double)(pops[j].size()))*popx[i].size());
		for (int g = 0; g < genestoconsider; ++g) {
			finalval += ranking[g].first;
		}

		F[i][j] = finalval;
	}


	return F;
}

//Problem-independent
std::vector<std::vector<double>> neighDist(int MAX_T, const std::vector<std::vector<double>> &F) {
	std::vector<std::vector<double>> dist(F[0].size(), std::vector<double>(2, 0.0));	//2 axes: 0-> how bad; 1-> how good
	std::vector<std::pair<double, int>> projectB(F[0].size(), std::make_pair(9999999999.9, 0)); //projection on the "best value obtained"-axis
	std::vector<std::pair<double, int>> projectW(F[0].size(), std::make_pair(-9999999999.9, 0)); //projection on the "worst value obtained"-axis


	for (int i = 0; i < F.size(); i++) {
		for (int j = 0; j < F[i].size(); j++) {
			if (F[i][j] < projectB[j].first) {
				projectB[j].first = F[i][j];
				projectB[j].second = j;
			}
			if (F[i][j] > projectW[j].first) {
				projectW[j].first = F[i][j];
				projectW[j].second = j;
			}
		}
	}
	std::sort(projectB.begin(), projectB.end());
	std::sort(projectW.begin(), projectW.end());

	//ensure theoretical extremes are kept in the same positions (0 and 1)
	if (projectB[0].second != 1) {
		for (int i = 1; i < projectB.size(); i++) {
			if (projectB[i].second == 1) {
				std::swap(projectB[0], projectB[i]);
				break;
			}
		}
	}
	if (projectW[F[0].size() - 1].second != 0) { 
		for (int i = projectW.size() - 2; i >= 0; i--) {
			if (projectW[i].second == 0) {
				std::swap(projectW[F[0].size() - 1], projectW[i]);
				break;
			}
		}
	}

	//calculate distance
	double BigM = 0.0;
	for (int i = 1; i < F[0].size() - 1; i++) {
		dist[projectB[i].second][0] = abs(projectB[i - 1].first - projectB[i].first)*
			abs(projectB[i + 1].first - projectB[i].first);
		dist[projectW[i].second][1] = abs(projectW[i - 1].first - projectW[i].first)*
			abs(projectW[i + 1].first - projectW[i].first);
		if (dist[projectB[i].second][0] > BigM) {
			BigM = dist[projectB[i].second][0];
		}
		if (dist[projectW[i].second][1] > BigM) {
			BigM = dist[projectW[i].second][1];
		}
	}
	//extremes get higher distance
	dist[projectB[0].second][0] = BigM + 1;
	dist[projectB[F[0].size() - 1].second][0] = BigM + 1;
	dist[projectW[0].second][1] = BigM + 1;
	dist[projectW[F[0].size() - 1].second][1] = BigM + 1;

	return dist;

}

//Problem-independent
double SolutionDecoder::soludecode(const int id, const std::vector< double >& chromosome, const std::vector<std::vector<double>> &F, const int solucriterion) const {
	int nscen = F[0].size();
	double myFitness = 0.0;
	double myFinalFitness = 0.0;


	if (solucriterion == 2) { //pessimistic criterion (worst-case)
		for (int s = 0; s < nscen; s++) {
			myFitness = F[id][s];
			if (s == 0) {
				myFinalFitness = myFitness;
			}
			else {
				if (myFitness > myFinalFitness) { 
					myFinalFitness = myFitness;
				}
			}
		}
	}
	else if (solucriterion == 3) { //optimistic criterion (best-case)
		for (int s = 0; s < nscen; s++) {
			myFitness = F[id][s];
			if (s == 0) {
				myFinalFitness = myFitness;
			}
			else {
				if (myFitness < myFinalFitness) { 
					myFinalFitness = myFitness;
				}
			}
		}
	}
	else if (solucriterion == 1) { //laplace criterion (average)
		for (int s = 0; s < nscen; s++) {
			myFitness += F[id][s];
		}
		myFinalFitness = myFitness / (double)(F[id].size());
	}


	return myFinalFitness;
}

//Problem-independent
double ScenarioDecoder::scendecode(const int id, const std::vector<double> &D) const {

	double myFinalFitness = 0.0;
	double maxneighdist = D[id];

	myFinalFitness = -maxneighdist;

	return myFinalFitness;

}