#include <iostream>
#include "SampleDecoder.h"
#include "MTRand.h"
#include "BRKGA.h"

int main(int argc, char* argv[]) {
	const unsigned solu_n = 100;		// size of solution chromosomes 
	const unsigned solu_p = 100;		// size of solution population 
	const double solu_pe = 0.20;		// fraction of solution population to be the elite-set
	const double solu_pm = 0.10;		// fraction of solution population to be replaced by mutants
	const double solu_rhoe = 0.70;		// probability that solution offspring inherit an allele from elite parent
	const unsigned scen_n = 100;		// size of scenario chromosomes 
	const unsigned scen_p = 30;			// size of scenario population 
	const double scen_pe = 0.20;		// fraction of scenario population to be the elite-set
	const double scen_pm = 0.10;		// fraction of scenario population to be replaced by mutants
	const double scen_rhoe = 0.70;		// probability that scenario offspring inherit an allele from elite parent

	const unsigned K = 1;				// number of independent populations
	const unsigned MAXT = 1;			// number of threads for parallel decoding
	
	const int solucriterion = 1;						//criterion for performace of solutions: 1- Laplace (average), 2- Pessimistic (worst-case), 3- Optimistic (best-case)
	const std::vector<double> extremeworst (scen_n,1);	//theoretical worst scenario (if not defined, this must be an empty vector)
	const std::vector<double> extremebest (scen_n,0);	//theoretical best scenario (if not defined, this must be an empty vector)
	
	SolutionDecoder soludecoder;			// initialize the solution decoder
	ScenarioDecoder scendecoder;		// initialize the scenario decoder

	
	const long unsigned rngSeed = 0;	// seed to the random number generator
	MTRand rng(rngSeed);				// initialize the random number generator
	
	// initialize the BRKGA-based heuristic
	BRKGA< SolutionDecoder, ScenarioDecoder, MTRand > algorithm(solu_n, solu_p, solu_pe, solu_pm, solu_rhoe,
		scen_n, scen_p, scen_pe, scen_pm, scen_rhoe, extremeworst, extremebest, soludecoder, scendecoder, rng, solucriterion, K, MAXT);
	
	unsigned generation = 0;		// current generation
	const unsigned X_INTVL = 100;	// exchange best individuals at every 100 generations
	const unsigned X_NUMBER = 2;	// exchange top 2 best
	const unsigned MAX_GENS = 1000;	// run for 1000 gens
	do {
		algorithm.evolve();	// evolve the population for one generation
		
		if((++generation) % X_INTVL == 0) {
			algorithm.exchangeEliteSolu(X_NUMBER);	// exchange top individuals (solutions)
			//algorithm.exchangeEliteScen(X_NUMBER);	// exchange top individuals (scenarios)
		}

		std::cout << generation << '\n';
	} while (generation < MAX_GENS);
	
	std::cout << "Best solution found has criterion value = "
	 		<< algorithm.getBestFitnessSolu() << std::endl;
	

	//Get final populations (scenarios and solutions)
	std::vector<std::vector<std::vector<double>>> scenarios(K);
	for (int i = 0; i < K; i++) {
		scenarios[i] = algorithm.getPopulationScen(i).getChromosomes();
	}

	std::vector<std::vector<std::vector<double>>> solutions(K);
	for (int k = 0; k < K; k++) {
		solutions[k] = algorithm.getPopulationSolu(k).getChromosomes();
	}

	return 0;
}